#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// decrypt the program rom (remove original IGS encryption)
static void pgm_decrypt_py2k2(unsigned char *rom, int length) // and ddp3
{
	int i;
	unsigned short *src = (unsigned short *)rom;

	static const unsigned char py2k2_tab[256] = {
		0x74, 0xe8, 0xa8, 0x64, 0x26, 0x44, 0xa6, 0x9a, 0xa5, 0x69, 0xa2, 0xd3, 0x6d, 0xba, 0xff, 0xf3,
		0xeb, 0x6e, 0xe3, 0x70, 0x72, 0x58, 0x27, 0xd9, 0xe4, 0x9f, 0x50, 0xa2, 0xdd, 0xce, 0x6e, 0xf6,
		0x44, 0x72, 0x0c, 0x7e, 0x4d, 0x41, 0x77, 0x2d, 0x00, 0xad, 0x1a, 0x5f, 0x6b, 0xc0, 0x1d, 0x4e,
		0x4c, 0x72, 0x62, 0x3c, 0x32, 0x28, 0x43, 0xf8, 0x9d, 0x52, 0x05, 0x7e, 0xd1, 0xee, 0x82, 0x61,
		0x3b, 0x3f, 0x77, 0xf3, 0x8f, 0x7e, 0x3f, 0xf1, 0xdf, 0x8f, 0x68, 0x43, 0xd7, 0x68, 0xdf, 0x19,
		0x87, 0xff, 0x74, 0xe5, 0x3f, 0x43, 0x8e, 0x80, 0x0f, 0x7e, 0xdb, 0x32, 0xe8, 0xd1, 0x66, 0x8f,
		0xbe, 0xe2, 0x33, 0x94, 0xc8, 0x32, 0x39, 0xfa, 0xf0, 0x43, 0xde, 0x84, 0x18, 0xd0, 0x6d, 0xd5,
		0x74, 0x98, 0xf8, 0x64, 0xcf, 0x84, 0xc6, 0xea, 0x55, 0x32, 0xe2, 0x38, 0xdd, 0xea, 0xfd, 0x6c,
		0xeb, 0x6e, 0xe3, 0x70, 0xae, 0x38, 0xc7, 0xd9, 0x54, 0x84, 0x10, 0xc1, 0xfd, 0x1e, 0x6e, 0x6d,
		0x37, 0xe0, 0x03, 0x9e, 0x06, 0x36, 0x68, 0x5b, 0xe3, 0xf6, 0x7f, 0x0b, 0x56, 0x79, 0xe0, 0xa8,
		0x98, 0x77, 0xc7, 0x2b, 0xa5, 0x79, 0xff, 0x2f, 0xca, 0x15, 0x71, 0x7e, 0x02, 0xbf, 0x87, 0xb7,
		0x7a, 0x8e, 0xe6, 0x64, 0x32, 0x62, 0x2a, 0xca, 0x23, 0x72, 0x87, 0xb5, 0x0c, 0x02, 0x4b, 0xee,
		0x44, 0x72, 0x9c, 0x7e, 0x5d, 0xc1, 0xa7, 0x1d, 0x30, 0x38, 0xda, 0xc9, 0x5b, 0xd0, 0x11, 0xf9,
		0xb1, 0x72, 0x6c, 0x04, 0x31, 0xc9, 0x50, 0x60, 0x6f, 0xc1, 0xf2, 0xae, 0x00, 0xf4, 0x5d, 0x66,
		0x43, 0x0e, 0x7a, 0xc3, 0x76, 0xae, 0x3c, 0xc2, 0xb7, 0xc9, 0x52, 0xf4, 0x74, 0x51, 0xaf, 0x12,
		0x19, 0xc6, 0x75, 0xe8, 0x6c, 0x54, 0x7e, 0x63, 0xdd, 0xae, 0x07, 0x5a, 0xb7, 0x00, 0xb5, 0x5e
	};

	for (i = 0; i < length/2; i++)
	{
		unsigned short x = src[i];

		if ((i & 0x040480) != 0x000080) x ^= 0x0001;
		if ((i & 0x084008) == 0x084008) x ^= 0x0002;
		if ((i & 0x000030) == 0x000010 && (i & 0x180000) != 0x080000) x ^= 0x0004;
		if ((i & 0x000042) != 0x000042) x ^= 0x0008;
		if ((i & 0x008100) == 0x008000) x ^= 0x0010;
		if ((i & 0x022004) != 0x000004) x ^= 0x0020;
		if ((i & 0x011800) != 0x010000) x ^= 0x0040;
		if ((i & 0x004820) == 0x004820) x ^= 0x0080;

		x ^= py2k2_tab[i & 0xff] << 8;

		src[i] = x;
	}
}

// remove ASIC27A copy protection
static void apply_patches(unsigned char *rom, int length)
{
	int i;
	unsigned short *src = (unsigned short*)rom;

	unsigned int rom_base = 0x100400; // rom location where the new code will reside
	unsigned int ram_base = 0x81e420; // ram location where the new code will store variables

	for (i = 0; i < length/2; i++)
	{
		// write data for command directly to ram
		if (src[i] == 0x33c0 && src[i+1] == 0x50 && src[i+2] == 0)
		{
			src[i+1] = ram_base >> 16;
			src[i+2] = ram_base & 0xffff;
		}

		// move.w D0, $500002, this is the command, intercept it here
		if (src[i] == 0x33c0 && src[i+1] == 0x50 && src[i+2] == 2)
		{
			src[i+0] = 0x4eb9;	// jsr
			src[i+1] = rom_base >> 16;
			src[i+2] = rom_base & 0xffff;
		}

		// read data back from the asic. point directly to where we store the results
		if ((src[i] & 0xf0ff) == 0x3039 && src[i+1] == 0x50 && src[i+2] <= 2)
		{
			src[i+1] = ram_base >> 16;
			src[i+2] = (ram_base + (src[i+2] ^ 2) + 8) & 0xffff;
		}

		// intercept write command value directly from a specific ram address to the asic
		if (src[i] == 0x33f9 && src[i+3] == 0x50 && src[i+4] == 0)
		{
			src[i+3] = ram_base >> 16;
			src[i+4] = ram_base & 0xffff;
		}

		// intercept write command directly from a specific ram address to the asic
		if (src[i] == 0x33f9 && src[i+3] == 0x50 && src[i+4] == 2)
		{
			src[i+0] = 0x4eb9; // jsr
			src[i+1] = rom_base >> 16;
			src[i+2] = (rom_base + 2) & 0xffff;
		}

		// 146508: 4279 0050 0004           clr.w   $500004.l
		if (src[i] == 0x4279 && src[i+1] == 0x50 && src[i+2] == 4) {
			src[i+0] = 0x4e71; // nop
			src[i+1] = 0x4e71;
			src[i+2] = 0x4e71;
		}
		
		// remove nvram magic number protection from black label versions (part 1)
		if (src[i] == 0x0cb9 && ((src[i+1] == 0x3698 && src[i+2] == 0x2136) || (src[i+1] == 0x9580 && src[i+2] == 0x4803)) && src[i+3] == 0x80 && src[i+4] == 0x3800)
		{
			src[i+5] = 0x4e71; // remove bne
			src[i+6] = 0x4e71;
		}

		// remove nvram magic number protection from black label versions (part 2)
		if (src[i] == 0x0cb9 && ((src[i+1] == 0x7634 && src[i+2] == 0x9621) || (src[i+1] == 0x2387 && src[i+2] == 0x9065)) && src[i+3] == 0x80 && src[i+4] == 0x3804)
		{
			src[i+5] = 0x4e71; // remove bne
			src[i+6] = 0x4e71;
		}
	}

	// load the file with our assembly
	{
		int t;
		FILE *fa;

		fa = fopen("ddp3_prot.bin", "rb");
		if (fa == NULL)
		{
			printf ("Could not find ddp3_prot.bin! Please put this in the same location as this program!\n");
			printf ("Your file will not work correctly.\n");
			return;
		}
		
		fseek (fa, 0, SEEK_END);
		length = ftell (fa);
		rewind (fa);

		fread (rom + 0x400, length, 1, fa);
		fclose (fa);

		// byteswap the patched in code
		for (i = 0; i < ((length + 1) & ~1); i+=2) {
			t = rom[0x400 + i];
			rom[0x400+i] = rom[0x400+i+1];
			rom[0x400+i+1] = t;
		}
	}
}

int main()
{
	FILE *fa;
	int length;
	char romname[128];
	unsigned char *src = (unsigned char*)malloc(0x200000);
	if (src == NULL) {
		printf ("Failed to allocate memory!\n");
		return 0;
	}

	printf ("Please type the name of the input file:\n");
	gets (romname);

	fa = fopen(romname, "rb");
	if (fa == NULL)
	{
		free (src);
		printf ("%s not found!\n", romname);
		return 0;
	}

	fseek (fa, 0, SEEK_END);

	length = ftell (fa);
	rewind (fa);

	if (length > 0x200000) length = 0x200000;
	if (length < 0x100000) {
		printf ("invalid rom size! %d\n", length);
		free (src);
		fclose (fa);
		return 0;
	}

	fread (src, length, 1, fa);
	fclose (fa);

	pgm_decrypt_py2k2(src, length);

	apply_patches(src, length);

	strcat (romname, "-deprotected.bin");

	fa = fopen(romname, "wb");
	fwrite (src, length, 1, fa);
	fclose (fa);

	free (src);

	return 1;
}
